/*
    jdkdrum, a simple linux command line drum machine
    Copyright (C) 1998  J.D. Koftinoff Software, Ltd. 
    http://www.jdkoftinoff.com/

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "jdkdrum.h"

long audiobuf_size;
volatile int audiobuf_current_playing;
long audiobuf_repeatpos[2];
signed short *audiobuf[2];
volatile int keep_playing=1;
volatile int play_count=0;

int audio_block_size;
int audio_fd;


int open_audio( void ) 
{
  int audio_format = AFMT_S16_LE;
  int audio_stereo = 1;
  int audio_speed = SAMPLE_RATE;
  
  audio_fd=open( AUDIO_DEVICE, O_WRONLY, 0 );
  
  if( audio_fd<0 )
  {
    perror( "Error opening audio device" );
    return -1;
  }
  
  if (ioctl( audio_fd, SNDCTL_DSP_SYNC, NULL) < 0)
  {
    perror( "Error syncing dsp" );
    return -1;
  }
  
  
  
  if( ioctl(audio_fd, SNDCTL_DSP_SETFMT, &audio_format ) == -1 
      || audio_format != AFMT_S16_LE )
  {
    perror( "Error selecting 16 bit audio format" );
    return -1;
  }
  
  
  if( ioctl(audio_fd, SNDCTL_DSP_STEREO, &audio_stereo )== -1
      || audio_stereo!=1 )
  {
    perror( "Error selecting stereo audio format" );
    return -1;
  }
  
  
  if( ioctl(audio_fd,SNDCTL_DSP_SPEED, &audio_speed)==-1 
      || audio_speed!=SAMPLE_RATE )
  {
    perror("Error selecting sample rate" );
    return -1;
  }
  
  if( ioctl(audio_fd,SNDCTL_DSP_GETBLKSIZE, &audio_block_size) < 0 )
  {
    perror( "Error getting DSP block size" );
    return -1;
  }
  
  return 0;
}



void close_audio(void) 
{
  close( audio_fd );
}



int write_audio( signed short *buf, int num_samples ) 
{
  write( audio_fd, buf, num_samples *2*2 );	
}



void *repeat_play_thread(void *f) 
{
  while(keep_playing)
  {
    write_audio( 
      audiobuf[audiobuf_current_playing], 
      audiobuf_repeatpos[audiobuf_current_playing] 
      );
    play_count++;
  }
  
  pthread_exit(0);
}



pthread_t start_repeat_play_thread(void) 
{
  pthread_t thr;
  
  /* Create the thread */
  if (pthread_create(&thr, 0, repeat_play_thread, 0))
  {		
    perror("Cannot create thread");
    return 0;
  }	
  return thr;
}



int merge_buffer( 
  signed short *audiobuf,
  int repeatpos,
  signed short *buf,
  int buflen,
  int pos, 
  float vol, 
  float pan
  ) 
{
  int i;
  float left_vol = (2.0-(pan+1.0))*0.5 * vol;
  float right_vol = (pan+1.0)*0.5 * vol;
  
  for( i=0; i<buflen; i++ )
  {			
    
    audiobuf[(pos*2)+0] += (signed short)((float)buf[i] * left_vol);
    audiobuf[(pos*2)+1] += (signed short)((float)buf[i] * right_vol);
    
    pos++;
    if( pos>repeatpos )
    {
      pos=0;
    }			
  }
  
  return 0;
}



int merge_read_file( 
  signed short *audiobuf,
  int repeatpos,
  const char *fname, 
  int pos, 
  float vol, 
  float pan
  ) 
{
  FILE *f = fopen(fname,"rb");
  if( f )
  {
    float left_vol = (2.0-(pan+1.0))*0.5 * vol;
    float right_vol = (pan+1.0)*0.5 * vol;
    
    while( !feof(f) )
    {			
      signed short val;
      
      if( fread(&val,2,1,f)!=1 )
        break;
      
      audiobuf[(pos*2)+0] += (signed short)((float)val * left_vol);
      audiobuf[(pos*2)+1] += (signed short)((float)val * right_vol);
      
      pos++;
      if( pos>repeatpos )
      {
        pos=0;
      }			
    }
    
    fclose(f);
    return 0;
  }
  
  return -1;
}



int merge_echo_read_file( 
  signed short *audiobuf,
  int repeatpos,
  const char *fname, 
  int pos, 
  float vol, 
  float pan,
  int echo_left_time,
  int echo_right_time,
  float echo_left_volume,
  float echo_right_volume
  ) 
{
  FILE *f = fopen(fname,"rb");
  if( f )
  {
    float left_vol = (2.0-(pan+1.0))*0.5 * vol;
    float right_vol = (pan+1.0)*0.5 * vol;
    
    while( !feof(f) )
    {			
      signed short val;
      
      if( fread(&val,2,1,f)!=1 )
        break;
      
      audiobuf[(((pos+echo_left_time)%repeatpos)*2)+0]
        += (signed short)((float)val * echo_left_volume * left_vol );
      
      audiobuf[(((pos+echo_right_time)%repeatpos)*2)+1]
        += (signed short)((float)val * echo_right_volume  * right_vol);
      
      audiobuf[(pos*2)+1] += (signed short)((float)val * right_vol );
      
      
      audiobuf[(pos*2)+0] += (signed short)((float)val * left_vol);
      audiobuf[(pos*2)+1] += (signed short)((float)val * right_vol);
      
      pos++;
      if( pos>repeatpos )
      {
        pos=0;
      }			
      
      
    }
    
    fclose(f);
    return 0;
  }
  
  return -1;
}



signed short * load_instrument( 
  const char *fname, 
  int *len 
  ) 
{
  FILE *f = fopen( fname, "rb" );
  if( f )
  {
    signed short *buf;
    size_t filelen;
    
    fseek(f,0,SEEK_END);
    
    filelen=ftell(f);
    fseek(f,0,SEEK_SET);
    
    buf=(signed short*)malloc(filelen);
    
    if( buf )
    {
      if( fread( buf, 1, filelen, f )!=filelen )
      {
        fclose(f);
        free(buf);
        return 0;
      }
      
    }
    
    fclose(f);
    
    *len = filelen/2;	
    
    return buf;
  }
  else
  {		
    return 0;
  }
  
} 


int merge_text_pattern(
  signed short *audiobuf,
  int repeatpos,
  const char *fname,
  int step_position[16],
  const char *pattern
  ) 
{
  int i;
  
  signed short *instrument_data;
  int instrument_data_len = 0;
  
  instrument_data = load_instrument( fname, &instrument_data_len );
  
  if( instrument_data )
  {		
    for( i=0; i<16; ++i )
    {
      float vol=0.0f;
      float pan=(float)(rand()%200) / 100.0 - 1.0;
      
      if( pattern[i]>='1' && pattern[i]<='9' )
      {
        vol = (pattern[i]-'0') * 0.1;
        merge_buffer( 
          audiobuf,
          repeatpos,
          instrument_data, 
          instrument_data_len, 
          step_position[i],
          vol,
          pan 							 
          );
        
      }
      
    }
    free(instrument_data);
    return 0;
  }
  else
  {
    fprintf( stderr, "error loading instrument '%s'\n", fname );		
  }
  
  
  return -1;
  
} 


int merge_multi_pattern(
  FILE *in
  ) 
{
  int i;
  int step_position[16];	
  float tempo_bpm = 135.0;
  float tempo_bps;
  int beats_per_measure=4;
  int repeatpos;
  signed short *audiobuf_to_use = audiobuf[audiobuf_current_playing^1];
  
  fscanf( in, "%f", &tempo_bpm );
  if( feof(in) )
    return -1;
  
  //fprintf( stderr, "%f\n", tempo_bpm );
  tempo_bps = tempo_bpm / 60;	
  repeatpos = (long)(((float)SAMPLE_RATE) / tempo_bps ) * beats_per_measure;
  
  
  for( i=0; i<16; ++i )
  {
    step_position[i] = (repeatpos*i/16);	
  }
  
  for( i=0; i<repeatpos*2; ++i )
  {
    audiobuf_to_use[i]=0;	
  }
  
  
  while(!feof(in))
  {		
    char fname[512];
    char pat[128];
    
    if( fscanf( in, "%510s", fname )==1 )
    {
      if( !strcmp( fname, "." ) )
      {
        break;
      }
      
      
      if( fscanf( in, "%125s", pat )==1 )
      {			
        
        //fprintf( stderr, "%32s  %16s\n", fname, pat );
        merge_text_pattern(
          audiobuf_to_use,
          repeatpos,
          fname,
          step_position,
          pat
          );
      }
      else
      {
        break;
      }					  		
    }
    else
    {
      break;
    }
    
  }
  
#if 0
  {
    // hack to wait for beginning of next measure
    int old_play_count = play_count;
    while( play_count==old_play_count )
    {
      usleep(100);
    }
  }
#endif	
  // TODO: race condition here needs fixing
  audiobuf_repeatpos[audiobuf_current_playing^1] = repeatpos;
  
  return 0;
} 


int main( int argc, char **argv ) 
{
  int i;
  void * retval;
  int generate=0;
  pthread_t thr;
  
  if( argc<1)
  {
    fprintf( stderr, "jdkdrum: \nBy Jeff Koftinoff <jeffk@turnercom.com> http://www.turnercom.com/jdk\n\n" );
    
    return 1;   
  }
  
  if( argc>1 )
  {		
    if( strcmp( argv[1], "-g" )==0 )
    {
      generate=1;
    }		
  }
  
  
  audiobuf_size = SAMPLE_RATE*8;	// 8 secs
  audiobuf[0] = (signed short *)malloc(audiobuf_size*2*2);	
  audiobuf[1] = (signed short *)malloc(audiobuf_size*2*2);
  
  
  
  
  for( i=0; i<audiobuf_size*2; ++i )
  {
    audiobuf[0][i] = 0;		
    audiobuf[1][i] = 0;				
  }
  
  
  if( !generate )
  {		
    if( open_audio()<0 )
    {
      return 1;	
    }		
    
    signal( SIGINT, SIG_IGN );		
    
    thr= start_repeat_play_thread();
  }
  
  
  
  while( !feof(stdin) )
  {
    
    if( merge_multi_pattern( stdin )>=0 )
    {	
      audiobuf_current_playing^=1;
      
      if( generate )
      {
        fwrite( 
          audiobuf[audiobuf_current_playing], 
          4, 
          audiobuf_repeatpos[audiobuf_current_playing], 
          stdout 
          );					
      }
      else
      {
        {
          // hack to wait for beginning of next measure
          int old_play_count = play_count;
          
          
          
          while( play_count==old_play_count )
          {
            usleep(100);
          }
        }
        
      }
      
    }
    
  }
  
  
  if( !generate )
  {				
    // tell thread to stop next measure
    keep_playing=0;
    
    // wait for thread to end.
    pthread_join(thr, &retval );	
    close_audio();
  }
  
  free( audiobuf[0] );
  free( audiobuf[1] );
  return 0;	
}




